from enum import Enum

class SPIClockPolarity:
    IDLE_LOW = 0x00
    IDLE_HIGH = 0x02

class SPIClockPhase:
    LEADING_EDGE = 0x00
    TRAILING_EDGE = 0x01

class SPIBus:
    def __init__(self):
        pass

    # Abstract (overriden) methods

    def _getClockSpeed(self):
        raise NotImplementedError()
    def _setClockSpeed(self, hz):
        raise NotImplementedError()
    def _getMode(self):
        raise NotImplementedError()
    def _setMode(self, clockPolarity = None, clockPhase = None):
        raise NotImplementedError()
    def _transfer(self, nbytes = None, buffer = None):
        raise NotImplementedError()

    # Exposed methods

    def getClockSpeed(self):
        return self._getClockSpeed()
    def setClockSpeed(self, hz):
        if (int(hz) <= 0):
            raise ValueError("Clock frequency has to be a positive integer")
        self._setClockSpeed(hz)
    def getMode(self):
        return self._getMode()
    def setMode(self, clockPolarity = None, clockPhase = None):
        if (clockPolarity is None) and (clockPhase is None):
            return True
        if not isinstance(clockPolarity, SPIClockPolarity):
            raise ValueError("Clock polarity has to be a SPIClockPolarity instance")
        if not isinstance(clockPhase, SPIClockPhase):
            raise ValueError("Clock phase has to be a SPIClockPhase instance")

        self._setMode(clockPolarity, clockPhase)
    def transfer(self, nbytes = None, buffer = None):
        if (nbytes is None) and (buffer is None):
            raise ValueError("Length of transfer is not determinable without nbytes or a TX buffer")

        if (nbytes is not None) and (buffer is not None):
            if len(buffer) != nbytes:
                raise ValueError(f"Supplied length {nbytes} differs from buffer length {len(buffer)}")

        if buffer is not None:
            nbytes = len(buffer)
            buffer = bytes(buffer)
            if not isinstance(buffer, bytes):
                raise ValueError("Buffer has to be a bytes instance ...")
        else:
            # Buffer does not exist ...
            buffer = bytes( [ 0 ] * nbytes )

        return self._transfer(nbytes, buffer)



import atexit

from enum import Enum
from .exceptions import CommunicationError_Timeout, CommunicationError_ProtocolViolation

class FunctionGeneratorModulation(Enum):
	NONE = 0
	ASK = 1
	FSK = 2
	PSK = 3
	TRIGGER = 4
	AM = 5
	FM = 6
	PM = 7

class FunctionGeneratorWaveform(Enum):
	SINE = 0
	SQUARE = 1
	RECTANGLE = 2
	TRAPEZOID = 3
	CMOS = 4
	ADJPULSE = 5
	DC = 6
	TRGL = 7
	RAMP = 8
	NEGRAMP = 9
	STAIRTRGL = 10
	STARSTEP = 11
	NEGSTAIR = 12
	POSEXP = 13
	NEGEXP = 14
	PFALLEXP = 15
	NFALLEXP = 16
	POSLOG = 17
	NEGLOG = 18
	PFALLLOG = 19
	NFALLLOG = 20
	PFULLWAV = 21
	NFULLWAV = 22
	PHALFWAV = 23
	NHALFWAV = 24
	SINCPULSE = 26
	IMPULSE = 27
	AM = 28
	FM = 29
	CHIRP = 30
	WHITENOISE = 32
	LORENTZPULSE = 33
	ECGSIMULATION = 34

class FunctionGenerator:
	def __init__(
		self,

		nchannels = None,
		freqrange = ( None, None ),
		amplituderange = ( None, None ),
		offsetrange = ( None, None ),

		arbitraryWaveforms = False,
		hasFrequencyCounter = False,

		arbitraryWaveformLength = ( None, None ),
		arbitraryWaveformMinMax = ( None, None ),
		arbitraryNormalizeInDriver = False,

		supportedWaveforms = [],
		supportedTriggerModes = [],
		supportedModulations = [],

		commandRetries = 3
	):
		if int(nchannels) < 1:
			raise ValueError("A function generator has to have at least one channel")
		if (not isinstance(freqrange, list)) and (not isinstance(freqrange, tuple)):
			raise ValueError("Frequency range has to be either a list or a tuple")
		if (not isinstance(amplituderange, list)) and (not isinstance(amplituderange, tuple)):
			raise ValueError("Amplitude range has to be either a list or a tuple")
		if (not isinstance(offsetrange, list)) and (not isinstance(offsetrange, tuple)):
			raise ValueError("Amplitude range has to be either a list or a tuple")
		if (not isinstance(arbitraryWaveformLength, list)) and (not isinstance(arbitraryWaveformLength, tuple)):
			raise ValueError("Arbitrary waveform length has to be either a 2-list or 2-tuple")
		if (not isinstance(arbitraryWaveformMinMax, list)) and (not isinstance(arbitraryWaveformMinMax, tuple)):
			raise ValueError("Arbitrary waveform length has to be either a 2-list or 2-tuple")

		if len(freqrange) != 2:
			raise ValueError("Frequency range has to be a 2-tuple or 2-list")
		if (not isinstance(freqrange[0], float)) and (not isinstance(freqrange[0], int)):
			raise ValueError("Frequency range has to be specified with floats or ints")
		if (not isinstance(freqrange[1], float)) and (not isinstance(freqrange[1], int)):
			raise ValueError("Frequency range has to be specified with floats or ints")

		if len(amplituderange) != 2:
			raise ValueError("Amplitude range has to be a 2-tuple or 2-list")
		if (not isinstance(amplituderange[0], float)) and (not isinstance(amplituderange[0], int)):
			raise ValueError("Amplitude range has to be specified with floats or ints")
		if (not isinstance(amplituderange[1], float)) and (not isinstance(amplituderange[1], int)):
			raise ValueError("Amplitude range has to be specified with floats or ints")

		if len(offsetrange) != 2:
			raise ValueError("Offset range has to be a 2-tuple or 2-list")
		if (not isinstance(offsetrange[0], float)) and (not isinstance(offsetrange[0], int)):
			raise ValueError("Offset range has to be specified with floats or ints")
		if (not isinstance(offsetrange[1], float)) and (not isinstance(offsetrange[1], int)):
			raise ValueError("Offset range has to be specified with floats or ints")

		if len(arbitraryWaveformLength) != 2:
			raise ValueError("Arbitrary waveform shape length has to be a 2-tuple or 2-list")
		if (not isinstance(arbitraryWaveformLength[0], float)) and (not isinstance(arbitraryWaveformLength[0], int)):
			raise ValueError("Arbitrary waveform length has to be specified with floats or ints")
		if (not isinstance(arbitraryWaveformLength[1], float)) and (not isinstance(arbitraryWaveformLength[1], int)):
			raise ValueError("Arbitrary waveform length has to be specified with floats or ints")

		if len(arbitraryWaveformMinMax) != 2:
			raise ValueError("Arbitrary waveform min/max has to be a 2-tuple or 2-list")
		if (not isinstance(arbitraryWaveformMinMax[0], float)) and (not isinstance(arbitraryWaveformMinMax[0], int)):
			raise ValueError("Arbitrary waveform min/max has to be specified with floats or ints")
		if (not isinstance(arbitraryWaveformMinMax[1], float)) and (not isinstance(arbitraryWaveformMinMax[1], int)):
			raise ValueError("Arbitrary waveform min/max has to be specified with floats or ints")

		if not isinstance(supportedWaveforms, list):
			raise ValueError("Supported waveforms has to be a list")
		if not isinstance(supportedTriggerModes, list):
			raise ValueError("Supported waveforms has to be a list")
		if not isinstance(supportedModulations, list):
			raise ValueError("Supported waveforms has to be a list")

		for sv in supportedWaveforms:
			if not isinstance(sv, FunctionGeneratorWaveform):
				raise ValueError("Waveform has to be a FunctionGeneratorWaveform instance")
		for sm in supportedModulations:
			if not isinstance(sm, FunctionGeneratorModulation):
				raise ValueError("Modulation has to be a instance of FunctionGeneratorModulation")

		self._nchannels = nchannels
		self._has_arb_waveforms = bool(arbitraryWaveforms)
		self._has_frq_counter = bool(hasFrequencyCounter)

		self._freqrange = freqrange
		self._amprange = amplituderange
		self._offsetrange = offsetrange
		self._arbitraryWaveformLength = arbitraryWaveformLength
		self._arbitraryWaveformMinMax = arbitraryWaveformMinMax
		self._arbitraryNormalizeInDriver = bool(arbitraryNormalizeInDriver)

		self._supported_waveforms = supportedWaveforms
		self._supported_modulations = supportedModulations
		self._supported_trigger_modes = supportedTriggerModes

		self._commandretries = commandRetries

	# Abstract methods implemented by particular device implementations

	# Connection and off
	def _connect(self):
		raise NotImplementedException()
	def _disconnect(self):
		raise NotImplementedException()
	def _off(self):
		raise NotImplementedException()

	# Identification and serial information
	def _id(self):
		raise NotImplementedException()
	def _serial(self):
		raise NotImplementedException()

	# Channel configuration
	def _set_channel_waveform(self, channel = None, waveform = None, arbitrary = None):
		raise NotImplementedException()
	def _get_channel_waveform(self, channel = None):
		raise NotImplementedException()
	def _set_channel_frequency(self, channel = None, frequency = None):
		raise NotImplementedException()
	def _get_channel_frequency(self, channel = None):
		raise NotImplementedException()
	def _set_channel_amplitude(self, channel = None, amplitude = None):
		raise NotImplementedException()
	def _get_channel_amplitude(self, channel = None):
		raise NotImplementedException()
	def _set_channel_offset(self, channel = None, offset = None):
		raise NotImplementedException()
	def _get_channel_offset(self, channel = None):
		raise NotImplementedException()
	def _set_channel_duty(self, channel = None, duty = None):
		raise NotImplementedException()
	def _get_channel_duty(self, channel = None):
		raise NotImplementedException()
	def _set_channel_phase(self, channel = None, phase = None):
		raise NotImplementedException()
	def _get_channel_phase(self, channel = None):
		raise NotImplementedException()
	def _set_channel_enabled(self, channel = None, enable = None):
		raise NotImplementedException()
	def _is_channel_enabled(self, channel = None):
		raise NotImplementedException()

	def _upload_waveform(self, slot, wavedata, normalize = False):
		raise NotImplementedException()

	# Public API:
	#
	# ToDo: Implement retry and readback ...

	def connect(self):
		if self._usesContext:
			raise ValueError("Cannot use connect on a context managed (with) object")

		retry_cnt = self._commandretries
		while True:
			try:
				res = self._connect()
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue
			break

		if not res:
			return res

		self._usedConnect = True
		return True

	def disconnect(self):
		if self._usesContext:
			raise ValueError("Cannot use connect on a context managed (with) object")
		if not self._usedConnect:
			return True

		self._disconnect()
		self._usedConnect = False
		return True

	def off(self):
		return self._off()

	def identify(self):
		retry_cnt = self._commandretries
		while True:
			try:
				id = self._id()
				return id
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise retry_cnt
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def serial(self):
		retry_cnt = self._commandretries
		while True:
			try:
				return self._serial()
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def set_channel_waveform(self, channel, waveform = None, arbitrary = None):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")
		if (waveform is None) and (arbitrary is None):
			raise ValueError("Either arbitrary waveform index or waveform has to be specified")
		if (waveform is not None) and (arbitrary is not None):
			raise ValueError("Only waveform or arbitrary waveform index has to be specified, not both")
		if (not isinstance(waveform, FunctionGeneratorWaveform)) and (waveform is not None):
			raise ValueError(f"Supplied waveform {waveform} is not an instance of FunctionGeneratorWaveform")
		if (waveform not in self._supported_waveforms) and (waveform is not None):
			raise ValueError(f"The waveform {waveform} is not supported by this particular device")
		if (arbitrary is not None) and (not self._has_arb_waveforms):
			raise ValueError(f"Arbitrary waveforms are not supported by this device")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._set_channel_waveform(channel, waveform = waveform, arbitrary = arbitrary)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue
			break

	def get_channel_waveform(self, channel):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._get_channel_waveform(channel)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def set_channel_frequency(self, channel, frequency):
		channel = int(channel)
		frequency = float(frequency)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		if (frequency < self._freqrange[0]) or (frequency > self._freqrange[1]):
			raise ValueError(f"Requested frequency {frequency}Hz is outside of supported range by this device ({self._freqrange[0]}Hz to {self._freqrange[1]}Hz)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._set_channel_frequency(channel, frequency)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def get_channel_frequency(self, channel):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._get_channel_frequency(channel)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def set_channel_amplitude(self, channel, amplitude):
		channel = int(channel)
		amplitude = float(amplitude)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")
		if (amplitude < self._amprange[0]) or (amplitude > self._amprange[1]):
			raise ValueError(f"Requested amplitude {amplitude}V is outside supported range from {self._amprange[0]}V to {self._amprange[1]}V")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._set_channel_amplitude(channel, amplitude)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def get_channel_amplitude(self, channel):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._get_channel_amplitude(channel)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def set_channel_offset(self, channel, offset):
		channel = int(channel)
		offset = float(offset)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")
		if (offset < self._offsetrange[0]) or (offset > self._offsetrange[1]):
			raise ValueError(f"Requested offset {offset}V is outside supported range from {self._offsetrange[0]}V to {self._offsetrange[1]}V")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._set_channel_offset(channel, offset)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def get_channel_offset(self, channel):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._get_channel_offset(channel)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def set_channel_duty(self, channel, duty):
		channel = int(channel)
		duty = float(duty)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")
		if (duty < 0) or (duty > 100.0):
			raise ValueError(f"Requested duty cycle {duty}% is outside supported range from 0% to 100%")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._set_channel_duty(channel, duty)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def get_channel_duty(self, channel):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._get_channel_duty(channel)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def set_channel_phase(self, channel, phase):
		channel = int(channel)
		phase = float(phase)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._set_channel_phase(channel, phase)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def get_channel_phase(self, channel):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._get_channel_phase(channel)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def set_channel_enabled(self, channel, enable):
		channel = int(channel)
		enable = bool(enable)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._set_channel_enabled(channel, enable)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def is_channel_enabled(self, channel):
		channel = int(channel)
		if (channel < 0) or (channel >= self._nchannels):
			raise ValueError(f"Supplied channel index is out of range ({self._nchannels} supported channels)")

		retry_cnt = self._commandretries
		while True:
			try:
				return self._is_channel_enabled(channel, enable)
			except (CommunicationError_Timeout, CommunicationError_ProtocolViolation) as e:
				if retry_cnt == 0:
					raise e
				if retry_cnt > 0:
					retry_cnt = retry_cnt - 1
				continue

	def upload_waveform(self, slot, wavedata, normalize = False):
		if not self._has_arb_waveforms:
			raise NotImplementedException("This device does not support arbitrary wavefunctions")

		try:
			ln = len(wavedata)
		except:
			raise ValueError("Wavedata has to be either a list, tuple or numpy array")

		if (len(wavedata) < self._arbitraryWaveformLength[0]) or (len(wavedata) > self._arbitraryWaveformLength[1]):
			raise ValueError("Waveform data has to be in range [{self._arbitraryWaveformLength[0]}; {self._arbitraryWaveformLength[1]}]")

		if normalize and not self._arbitraryNormalizeInDriver:
			mi = min(wavedata)
			mx = max(wavedata)
			rng = mx - mi
			if rng == 0:
				rng = 1
			for i in range(len(wavedata)):
				wavedata[i] = ((wavedata[i] - mi) / rng) * (self._arbitraryWaveformMinMax[1] - self._arbitraryWaveformMinMax[0]) + self._arbitraryWaveformMinMax[0]

			if (min(wavedata) < self._arbitraryWaveformMinMax[0]) or (max(wavedata) > self._arbitraryWaveformMinMax[1]):
				raise ValueError(f"The DDS DAC requires values from {self._arbitraryWaveformMinMax[0]} to {self._arbitraryWaveformMinMax[1]}")

		self._upload_waveform(slot, wavedata, normalize)



import atexit

from enum import Enum

class PowerSupplyLimit(Enum):
	NONE = 0
	VOLTAGE = 1
	CURRENT = 2

class PowerSupply:
	def __init__(
		self,
		nChannels = None,
		vrange = None,
		arange = None,
		prange = None,
		capableVLimit = True,
		capableALimit = True,
		capableMeasureV = True,
		capableMeasureA = True,
		capableOnOff = True
	):
		if not isinstance(nChannels, int):
			raise ValueError("Number of channels required")
		if nChannels < 1:
			raise ValueError("Power supply has to support at least one channel")
		if (vrange is None) or (arange is None) or (prange is None):
			raise ValueError("Voltage, current and power range has to be supplied")
		if (not isinstance(vrange, tuple)) or (not isinstance(arange, tuple)) or (not isinstance(prange, tuple)):
			raise ValueError("Voltage, current and power ranges have to be tuples")
		if (len(vrange) != 3) or (len(arange) != 3) or (len(prange) != 3):
			raise ValueError("Ranges have to supply minimum, maximum and step size")

		self._usesContext = False
		self._usedConnect = False

		self._nchannels = nChannels
		self._vrange = vrange
		self._arange = arange
		self._prange = prange
		self._capabilities = {
			'vlimit' : capableVLimit,
			'alimit' : capableALimit,
			'measureV' : capableMeasureV,
			'measureA' : capableMeasureA,
			'onoff' : capableOnOff
		}

		# Note that implementations have to update _setValues themselves!!
		self._setValues = []
		for i in range(nChannels):
			self._setValues.append({
				'onoff' : False,
				'volts' : 0.0,
				'amps' : 0.0
			})

		# We ensure calling "off" whenever the process gets terminated ...
		atexit.register(self._exitOff)

	def _exitOff(self):
		if self._isConnected():
			self.off()

	def _setChannelEnable(self, enable, channel):
		raise NotImplementedError()
	def _setVoltage(self, voltage, channel):
		raise NotImplementedError()
	def _setCurrent(self, current, channel):
		raise NotImplementedError()
	def _getVoltage(self, channel):
		raise NotImplementedError()
	def _getCurrent(self, channel):
		raise NotImplementedError()
	def _off(self):
		raise NotImplementedError()
	def _getLimitMode(self, channel):
		raise NotImplementedError()
	def _isConnected(self):
		raise NotImplementedError()
	def _connect(self):
		raise NotImplementedError()
	def _disconnect(self):
		raise NotImplementedError()

	def setChannelEnable(self, enable, channel = 1):
		if not isinstance(channel, int):
			raise ValueError("Channel has to be an integer number")
		if (channel < 1) or (channel > self._nchannels):
			raise ValueError("Channel {} is out of range (valid channel range is 1 to {})".format(channel, self._nchannels))
		if not isinstance(enable, bool):
			raise ValueError("Enable flag has to be True or False")
		if not self._capabilities['onoff']:
			raise TypeError("This device does not support on/off switching")

		return self._setChannelEnable(enable, channel)

	def setVoltage(self, voltage, channel = 1):
		if not isinstance(channel, int):
			raise ValueError("Channel has to be an integer number")
		if (channel < 1) or (channel > self._nchannels):
			raise ValueError("Channel {} is out of range (valid channel range is 1 to {})".format(channel, self._nchannels))
		if not (isinstance(voltage, float) or isinstance(voltage, int)):
			raise ValueError("Voltage has to be an integer in range {} to {}".format(vrange[0], vrange[1]))

		pTarget = abs(voltage * self._setValues[channel-1]['amps'])
		if (pTarget < self._prange[0]) or (pTarget > self._prange[1]):
			raise ValueError("Selected voltage lies out of supported power range ({} to {} watts)".format(self._prange[0], self._prange[1]))

		if self._setVoltage(voltage, channel):
			self._setValues[channel-1]['volts'] = voltage

	def setCurrent(self, current, channel = 1):
		if not isinstance(channel, int):
			raise ValueError("Channel has to be an integer number")
		if (channel < 1) or (channel > self._nchannels):
			raise ValueError("Channel {} is out of range (valid channel range is 1 to {})".format(channel, self._nchannels))
		if not (isinstance(current, float) or isinstance(current, int)):
			raise ValueError("Current has to be an integer in range {} to {}".format(arange[0], arange[1]))

		pTarget = abs(current * self._setValues[channel-1]['volts'])
		if (pTarget < self._prange[0]) or (pTarget > self._prange[1]):
			raise ValueError("Selected current lies out of supported power range ({} to {} watts)".format(self._prange[0], self._prange[1]))

		if self._setCurrent(current, channel):
			self._setValues[channel-1]['amps'] = current

	def getVoltage(self, channel = 1):
		if not isinstance(channel, int):
			raise ValueError("Channel has to be an integer number")
		if (channel < 1) or (channel > self._nchannels):
			raise ValueError("Channel {} is out of range (valid channel range is 1 to {})".format(channel, self._nchannels))

		measVolts = None
		if self._capabilities['measureV']:
			measVolts = self._getVoltage(channel)

		return measVolts, self._setValues[channel-1]['volts']

	def getCurrent(self, channel = 1):
		if not isinstance(channel, int):
			raise ValueError("Channel has to be an integer number")
		if (channel < 1) or (channel > self._nchannels):
			raise ValueError("Channel {} is out of range (valid channel range is 1 to {})".format(channel, self._nchannels))

		measCurrent = None
		if self._capabilities['measureA']:
			measCurrent = self._getCurrent(channel)

		return measCurrent, self._setValues[channel-1]['amps']

	def off(self):
		return self._off()

	def getLimitMode(self, channel = 1):
		if not isinstance(channel, int):
			raise ValueError("Channel has to be an integer number")
		if (channel < 1) or (channel > self._nchannels):
			raise ValueError("Channel {} is out of range (valid channel range is 1 to {})".format(channel, self._nchannels))

		return self._getLimitMode(channel)

	# Imperative connect & disconnect methods

	def connect(self):
		if self._usesContext:
			raise ValueError("Cannot use connect on a context managed (with) object")

		res = self._connect()
		if not res:
			return res

		self._usedConnect = True
		return True

	def disconnect(self):
		if self._usesContext:
			raise ValueError("Cannot use connect on a context managed (with) object")
		if not self._usedConnect:
			return True

		self._disconnect()
		self._usedConnect = False
		return True



from .exceptions import *
from .functiongenerator import FunctionGeneratorModulation, FunctionGeneratorWaveform
from .oscilloscope import OscilloscopeRunMode, OscilloscopeSweepMode, OscilloscopeTriggerMode, OscilloscopeTimebaseMode, OscilloscopeCouplingMode
from .powersupply import PowerSupplyLimit



import atexit

from enum import Enum

class PressureGaugeUnit(Enum):
    MBAR = 0
    PASCAL = 1
    TORR = 2

class PressureGauge:
    def __init__(
        self,
        measurementRange = ( None, None ),
        hasDegas = False,
        deviceSupportedUnits = [ PressureGaugeUnit.MBAR ],
        debug = False
    ):
        if not isinstance(measurementRange, tuple) and not isinstance(measurementRange, list):
            raise ValueError("Measurement range has to be tuple or list with two components")
        if len(measurementRange) != 2:
            raise ValueError("Measurement range can only have two component")
        if measurementRange[0] > measurementRange[1]:
            raise ValueError("Measurement range minimum is larger than maximum")
        if len(deviceSupportedUnits) < 1:
            raise ValueError("Measurement units of device are unknown")

        self._debug = debug

        self._usesContext = False
        self._usedConnect = False

        self._supportedUnits = deviceSupportedUnits
        self._hasDegas = hasDegas

    # Overriden methods

    def _get_serials(self):
        return None
    def _get_versions(self):
        raise NotImplementedError()

    def _get_device_type(self):
        raise NotImplementedError()

    def _degas(self, degasOn):
        raise NotImplementedError()

    def _get_pressure(self):
        raise NotImplementedError()

    def _get_unit(self):
        raise NotImplementedError()

    def _set_unit(self, unit):
        raise NotImplementedError()

    # Public methods

    def get_pressure(self, unit = PressureGaugeUnit.MBAR):
        if not isinstance(unit, PressureGaugeUnit):
            raise ValueError("Unit has to be one the supported pressuage gauge units")

        resp = self._get_pressure()

        if resp is None:
            return None

        if unit in resp['value']:
            return resp['value'][unit]

        # Check if we can perform the conversion from MBAR to the given unit
        if unit == PressureGaugeUnit.MBAR:
            return resp['mbar']
        elif unit == PressureGaugeUnit.TORR:
            return resp['mbar'] * 0.750062
        elif unit == PressureGaugeUnit.PASCAL:
            return resp['mbar'] * 100.0
        else:
            raise ValueError(f"Unknown unit {unit}")

    def get_device_serial(self):
        raise NotImplementedError()

    def get_unit(self):
        return self._get_unit()

    def set_unit(self, unit = PressureGaugeUnit.MBAR):
        if not isinstance(unit, PressureGaugeUnit):
            raise ValueError(f"Unit {unit} is not instance of PressureGaugeUnit")
        if unit not in self._supportedUnits:
            raise ValueError(f"Unit {unit} not supported by device (supported: {self._supportedUnits})")
        return self._set_unit(unit)

    def degas(self, degasOn = True):
        if not isinstance(degasOn, bool):
            raise ValueError("Enable or disable parameter has to be a boolean")

        if self._hasDegas:
            self._degas(degasOn)
        else:
            raise NotImplementedError("Device does not support degas")

    def connect(self):
        if self._usesContext:
            raise ValueError("Cannot use connect on a context managed (with) object")

        res = self._connect()
        if not res:
            return res

        self._usedConnect = True
        return True

    def disconnect(self):
        if self._usesContext:
            raise ValueError("Cannot use connect on a context managed (with) object")
        if not self._usedConnect:
            return True

        self._disconnect()
        self._usedConnect = False
        return True



import atexit

from enum import Enum

class GCodeDevice:
	def __init__(
		self,

		name = "GCODEDEVICE",

		supportedGCodes = [ ],
		usesThread = False,

		echoConsole = False,
		hasEndstops = True,

		useMetric = True,
		useAbsolute = True
	):
		self._name = name

		self._usesContext = False
		self._usedConnect = False

		self._supported_gcodes = supportedGCodes
		self._hasEndstops = hasEndstops

		self._echo_console = False

		self._cb_echo = None

		self._use_metric = useMetric
		self._use_absolute = useAbsolute

	# Overriden methods

	def _cmd_getEndstopState(self, sync = False):
		raise NotImplementedError("Getting endstop state is currently not implemented")
	def _cmd_set_metric(self, sync = True):
		raise NotImplementedError("Setting metric units is currently not implemented")
	def _cmd_set_imperial(self, sync = True):
		raise NotImplementedError("Setting imperial units is currently not implemented")
	def _cmd_set_absolute(self, sync = True):
		raise NotImplementedError("Setting absolut positioning is currently not implemented")
	def _cmd_set_relative(self, sync = True):
		raise NotImplementedError("Setting relative positioning is currently not implemented")
	def _cmd_move(self, X = None, Y = None, Z = None, feedrate = None, sync = False, travelOnly = False):
		raise NotImplementedError("Linear movement not implemented")
	def _cmd_home(self, homeXY = True, homeZ = True, sync = True):
		raise NotImplementedError("Homing not implemented")
	def _cmd_disable_steppers(self, X = True, Y = True, Z = True, sync = True):
		raise NotImplementedError("Disabling steppers not implemented")

	# Exposed API:
	def get_endstops(self, sync = False):
		if self._hasEndstops:
			return self._cmd_getEndstopState(sync = sync)
		else:
			return { }

	def move(self, X = None, Y = None, Z = None, feedrate = None, sync = False, travelOnly = False):
		return self._cmd_move(X = X, Y = Y, Z = Z, feedrate = feedrate, sync = sync, travelOnly = travelOnly)

	def home(self, homeXY = True, homeZ = True, sync = False):
		return self._cmd_home(homeXY = homeXY, homeZ = homeZ, sync = sync)

	def positions_relative(self, sync = False):
		return self._cmd_set_relative(sync)

	def positions_absolute(self, sync = False):
		return self._cmd_set_absolute(sync)

	def set_metric(self, sync = True):
		return self._cmd_set_metric(sync)

	def set_imperial(self, sync = True):
		return self._cmd_set_imperial(sync)

	def disable_steppers(self, X = True, Y = True, Z = True, sync = True):
		return self._cmd_disable_steppers(X = X, Y = Y, Z = Z, sync = sync)

	# Event handling

	def _got_echo_message(self, msg):
		# We received an echo message that we might want to display on the console (depending on our configuration)
		if self._echo_console:
			print(f"[{self._name}] {msg}")

	# Imperative connection handling

	def connect(self):
		if self._usesContext:
			raise ValueError("Cannot use connect on a context managed (with) object")
		res = self._connect()
		if not res:
			return res
		self._usedConnect = True
		return True

	def disconnect(self):
		if self._usesContext:
			raise ValueError("Cannot use connect on a context managed object (with)")
		if not self._usedConnect:
			return True

		self._disconnect()
		self._usedConnect = False
		return True



# A simple (synchronizing) serial ringbuffer used in some
# lab device implementations.

# This is a simple helper class

import threading
from collections import deque

class SerialRingBuffer:
    def __init__(self, bufferSize = 512):
        self.bufferSize = bufferSize
        self.buffer = [ ]
        for i in range(512):
            self.buffer.append(None)
        self.head = 0
        self.tail = 0
        self.lock = threading.Lock()

    def isAvailable(self, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        if self.head == self.tail:
            if blocking:
                self.lock.release()
            return False
        else:
            if blocking:
                self.lock.release()
            return True

    def available(self, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        res = 0
        if self.head >= self.tail:
            res = self.head - self.tail;
        else:
            res = self.bufferSize - self.tail + self.head
        if blocking:
            self.lock.release()
        return res

    def discard(self, len, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        avail = self.available(blocking = False)
        if avail < len:
            self.tail = (self.tail + avail) % self.bufferSize
        else:
            self.tail = (self.tail + len) % self.bufferSize
        if blocking:
            self.lock.release()
        return None

    def peek(self, distance, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        if distance >= self.available(blocking = False):
            if blocking:
                self.lock.release()
            return None
        else:
            res = self.buffer[(self.tail + distance) % self.bufferSize]
            if blocking:
                self.lock.release()
            return res

    def remainingCapacity(self, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        res = self.bufferSize - self.available(blocking = False) - 1
        if blocking:
            self.lock.release()
        return res

    def capacity(self, *ignore, blocking = True):
        return self.bufferSize

    def push(self, data, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        if isinstance(data, list):
            if self.remainingCapacity(blocking = False) <  len(data):
                # Raise error ... ToDo
                if blocking:
                    self.lock.release()
                return
            else:
                for c in data:
                    self.push(c, blocking = False)
        else:
            if self.remainingCapacity(blocking = False) < 1:
                # Raise error ... ToDo
                if blocking:
                    self.lock.release()
                return
            self.buffer[self.head] = data
            self.head = (self.head + 1) % self.bufferSize
        self.lock.release()

    def pop(self, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        ret = None
        if self.head != self.tail:
            ret = self.buffer[self.tail]
            self.tail = (self.tail + 1) % self.bufferSize
        if blocking:
            self.lock.release()
        return ret

    def read(self, len, *ignore, blocking = True):
        if blocking:
            self.lock.acquire()
        if self.available(blocking = False) < len:
            # ToDo Raise exception
            if blocking:
                self.lock.release()
            return None
        ret = [ ]
        for i in range(len):
            ret.append(self.buffer[(self.tail + i) % self.bufferSize])
        self.tail = (self.tail + len) % self.bufferSize
        if blocking:
            self.lock.release()
        return ret



from enum import Enum

class SIUNIT(Enum):
    ONE     = 0
    DECA    = 1
    HECTO   = 2
    KILO    = 3
    MEGA    = 4
    GIGA    = 5
    TERA    = 6
    PETA    = 7
    EXA     = 8
    ZETTA   = 9
    YOTTA   = 10
    RONNA   = 11
    QUETTA  = 12
    DECI    = 13
    CENTI   = 14
    MILLI   = 15
    MICRO   = 16
    NANO    = 17
    PICO    = 18
    FEMTO   = 19
    ATTO    = 20
    ZEPTO   = 21
    YOCTO   = 22
    RONTO   = 23
    QUECTO  = 24

class SiUtils:
    def __init__(self):
        self._tbl = {
                SIUNIT.ONE    : { 'factor' : 1    , 'symbol' : ''  , 'name' : 'one'    },

                SIUNIT.DECA   : { 'factor' : 1e1  , 'symbol' : 'da', 'name' : 'deca'   },
                SIUNIT.HECTO  : { 'factor' : 1e2  , 'symbol' : 'h' , 'name' : 'hecto'  },
                SIUNIT.KILO   : { 'factor' : 1e3  , 'symbol' : 'k' , 'name' : 'kilo'   },
                SIUNIT.MEGA   : { 'factor' : 1e6  , 'symbol' : 'M' , 'name' : 'mega'   },
                SIUNIT.GIGA   : { 'factor' : 1e9  , 'symbol' : 'G' , 'name' : 'giga'   },
                SIUNIT.TERA   : { 'factor' : 1e12 , 'symbol' : 'T' , 'name' : 'tera'   },
                SIUNIT.PETA   : { 'factor' : 1e15 , 'symbol' : 'P' , 'name' : 'peta'   },
                SIUNIT.EXA    : { 'factor' : 1e18 , 'symbol' : 'E' , 'name' : 'exa'    },
                SIUNIT.ZETTA  : { 'factor' : 1e21 , 'symbol' : 'Z' , 'name' : 'zetta'  },
                SIUNIT.YOTTA  : { 'factor' : 1e24 , 'symbol' : 'Y' , 'name' : 'yotta'  },
                SIUNIT.RONNA  : { 'factor' : 1e27 , 'symbol' : 'R' , 'name' : 'ronna'  },
                SIUNIT.QUETTA : { 'factor' : 1e30 , 'symbol' : 'Q' , 'name' : 'quetta' },

                SIUNIT.DECI   : { 'factor' : 1e-1 , 'symbol' : 'd' , 'name' : 'deci'   },
                SIUNIT.CENTI  : { 'factor' : 1e-2 , 'symbol' : 'c' , 'name' : 'centi'  },
                SIUNIT.MILLI  : { 'factor' : 1e-3 , 'symbol' : 'm' , 'name' : 'milli'  },
                SIUNIT.MICRO  : { 'factor' : 1e-6 , 'symbol' : 'u' , 'name' : 'micro'  },
                SIUNIT.NANO   : { 'factor' : 1e-9 , 'symbol' : 'n' , 'name' : 'nano'   },
                SIUNIT.PICO   : { 'factor' : 1e-12, 'symbol' : 'p' , 'name' : 'pico'   },
                SIUNIT.FEMTO  : { 'factor' : 1e-15, 'symbol' : 'f' , 'name' : 'femto'  },
                SIUNIT.ATTO   : { 'factor' : 1e-18, 'symbol' : 'a' , 'name' : 'atto'   },
                SIUNIT.ZEPTO  : { 'factor' : 1e-21, 'symbol' : 'z' , 'name' : 'zepto'  },
                SIUNIT.YOCTO  : { 'factor' : 1e-24, 'symbol' : 'y' , 'name' : 'yocto'  },
                SIUNIT.RONTO  : { 'factor' : 1e-27, 'symbol' : 'r' , 'name' : 'ronto'  },
                SIUNIT.QUECTO : { 'factor' : 1e-30, 'symbol' : 'q' , 'name' : 'quecto' }
        }

    def get_unit(self, unitname):
        if unitname is None:
            return SIUNIT.ONE

        if isinstance(unitname, SIUNIT):
            return unitname
        try:
            raise ValueError("Getting via numeric value is currently not supported")
        except:
            # Search by name
            strname = str(unitname).lower().strip()
            for unit in self._tbl:
                if self._tbl[unit]['name'] == strname:
                    return unit
            return None

    def get_factor(self, unit):
        un = self.get_unit(unit)
        if un is None:
            return 1
        return self._tbl[un]['factor']

    def get_symbol(self, unit):
        un = self.get_unit(unit)
        if un is None:
            return ""
        return self._tbl[un]['symbol']

    def get_name(self, unit):
        un = self.get_unit(unit)
        if un is None:
            return ""
        return self._tbl[un]['name']

    def get_closest(self, value):
        absval = value
        if absval < 0:
            absval = absval * -1.0

        prevunit = SIUNIT.ONE

        for un in self._tbl:
            if (absval < 1) and (self._tbl[un]['factor'] >= 1):
                continue

            if absval < 1:
                if absval > self._tbl[un]['factor']:
                    return un
            else:
                if absval < self._tbl[un]['factor']:
                    return prevunit
                prevunit = un

        if absval < 1:
            return SIUNIT.QUECTO
        else:
            return SIUNIT.QUETTA

    def get_str(self, value, unit = ""):
        unt = self.get_closest(value)
        factor, symb = self.get_factor(unt), self.get_symbol(unt)

        s = f"{value / factor:.2f} {symb}{unit}".strip()
        return s



class CommunicationError(Exception):
    pass

class CommunicationError_ProtocolViolation(CommunicationError):
    pass

class CommunicationError_Timeout(CommunicationError):
    pass

class CommunicationError_NotConnected(CommunicationError):
    pass


# Work in progress
#
# Base class for oscilloscopes

import atexit
import numpy as np

from enum import Enum

class OscilloscopeRunMode(Enum):
    RUN = 0,
    STOP = 1
    SINGLE = 2

class OscilloscopeSweepMode(Enum):
    AUTO = 0
    NORMAL = 1
    SINGLE = 2

    @classmethod
    def has_value(cls, v):
        return v in cls._value2member_map_

class OscilloscopeTriggerMode(Enum):
    EDGE = 0
    PULSE = 1
    SLOPE = 2

    @classmethod
    def has_value(cls, v):
        return v in cls._value2member_map_

class OscilloscopeTimebaseMode(Enum):
    MAIN = 0
    XY = 1
    ROLL = 2

    @classmethod
    def has_value(cls, v):
        return v in cls._value2member_map_

class OscilloscopeCouplingMode(Enum):
    DC = 0
    AC = 1
    GND = 2

    @classmethod
    def has_value(cls, v):
        return v in cls._value2member_map_

class Oscilloscope:
    def __init__(
        self,

        nChannels = None,

        supportedSweepModes = [ ],
        supportedTriggerModes = [ ],
        supportedTimebaseModes = [ ],
        supportedRunModes = [ ],
        supportedChannelCouplingModes = [],

        triggerForceSupported = False,

        timebaseScale = ( None, None ),
        voltageScale = ( None, None )
    ):
        if not isinstance(nChannels, int):
            raise ValueError("Channel count has to be an integer")
        if nChannels < 1:
            raise ValueError("Channel count has to be a positive integer")
        if not isinstance(supportedSweepModes, list) and not isinstance(supportedSweepModes, tuple):
            raise ValueError("Sweep modes have to be either list or tuples")
        if not isinstance(supportedTriggerModes, list) and not isinstance(supportedTriggerModes, tuple):
            raise ValueError("Sweep modes have to be either list or tuples")
        if not isinstance(supportedTimebaseModes, list) and not isinstance(supportedTimebaseModes, tuple):
            raise ValueError("Sweep modes have to be either list or tuples")
        if not isinstance(supportedRunModes, list) and not isinstance(supportedRunModes, tuple):
            raise ValueError("Sweep modes have to be either list or tuples")
        if not (isinstance(timebaseScale, tuple) or isinstance(timebaseScale, list)):
            raise ValueError("Timebase scale has to be supplied as list of tuple")
        if not (isinstance(supportedChannelCouplingModes, tuple) or isinstance(supportedChannelCouplingModes, list)):
            raise ValueError("Supported channel coupling modes have to be supplied as list of tuple")
        if not (isinstance(voltageScale, tuple) or isinstance(voltageScale, list)):
            raise ValueError("Voltage scale has to be a tuple or list")

        for rm in supportedRunModes:
            if not isinstance(rm, OscilloscopeRunMode):
                raise ValueError(f"Run mode {rm} not known by labdevs library")
        for rm in supportedSweepModes:
            if not isinstance(rm, OscilloscopeSweepMode):
                raise ValueError(f"Sweep mode {rm} not known by labdevs library")
        for tm in supportedTriggerModes:
            if not isinstance(tm, OscilloscopeTriggerMode):
                raise ValueError(f"Trigger mode {tm} not known by labdevs library")
        for tbm in supportedTimebaseModes:
            if not isinstance(tbm, OscilloscopeTimebaseMode):
                raise ValueError(f"Timebase mode {tbm} not known by labdevs library")
        for chc in supportedChannelCouplingModes:
            if not isinstance(chc, OscilloscopeCouplingMode):
                raise ValueError(f"Coupling mode {chc} is not known by labdevs library")

        if not isinstance(timebaseScale[0], float) and not isinstance(timebaseScale[0], int):
            raise ValueError("Timebase minima and maxima have to be floating point numbers in seconds")
        if not isinstance(timebaseScale[1], float) and not isinstance(timebaseScale[1], int):
            raise ValueError("Timebase minima and maxima have to be floating point numbers in seconds")

        if not isinstance(voltageScale[0], float) and not isinstance(voltageScale[0], int):
            raise ValueError("Voltage scale minimum has to be either float or integer")
        if not isinstance(voltageScale[1], float) and not isinstance(voltageScale[1], int):
            raise ValueError("Voltage scale minimum has to be either float or integer")

        self._usesContext = False
        self._usedConnect = False

        self._nchannels = nChannels
        self._supportedSweepModes = supportedSweepModes
        self._supportedTriggerModes = supportedTriggerModes
        self._supportedTimebaseModes = supportedTimebaseModes
        self._supportedRunModes = supportedRunModes
        self._trigger_force_supported = triggerForceSupported
        self._supportedChannelCouplingModes = supportedChannelCouplingModes

        self._timebase_scale = timebaseScale
        self._voltage_scale = voltageScale

        atexit.register(self._exitOff)

    def _exitOff(self):
        if self._isConnected():
            self.off()


    # Overriden "protected" methods

    def _off(self):
        raise NotImplementedError()
    def _set_trigger_mode(self, mode):
        raise NotImplementedError()
    def _get_trigger_mode(self):
        raise NotImplementedError()
    def _set_sweep_mode(self, mode):
        raise NotImplementedError()
    def _get_sweep_mode(self):
        raise NotImplementedError()
    def _set_run_mode(self, mode):
        raise NotImplementedError()
    def _get_run_mode(self):
        raise NotImplementedError()

    def _set_timebase_mode(self, mode):
        raise NotImplementedError()
    def _get_timebase_mode(self):
        raise NotImplementedError()
    def _set_timebase_scale(self, scale):
        raise NotImplementedError()
    def _get_timebase_scale(self):
        raise NotImplementedError()

    def _identify(self):
        raise NotImplementedError()
    def _isConnected(self):
        raise NotImplementedError()

    def _set_channel_enable(self, channel, enabled):
        raise NotImplementedError()
    def _is_channel_enabled(self, channel):
        raise NotImplementedError()
    def _set_channel_coupling(self, channel, mode):
        raise NotImplementedError()
    def _get_channel_coupling(self, channel):
        raise NotImplementedError()
    def _set_channel_probe_ratio(self, channel, ratio):
        raise NotImplementedError()
    def _get_channel_probe_ratio(self, channel):
        raise NotImplementedError()
    def _set_channel_scale(self, channel, scale):
        raise NotImplementedError()
    def _get_channel_scale(self, channel):
        raise NotImplementedError()

    def _query_waveform(self, channel, stats = None):
        raise NotImplementedError()

    # Public API

    def set_channel_enable(self, channel, enabled):
        if not isinstance(channel, int):
            raise ValueError(f"Channel {channel} is not an integer")
        if (channel < 0) or (channel > self._nchannels):
            raise ValueError(f"Channel {channel} is out of range from 0 to {self._nchannels}")
        if not isinstance(enabled, bool):
            raise ValueError(f"{enabled} is not a boolean argument")

        self._set_channel_enable(channel, enabled)

    def is_channel_enabled(self, channel):
        if not isinstance(channel, int):
            raise ValueError(f"Channel {channel} is not an integer")
        if (channel < 0) or (channel > self._nchannels):
            raise ValueError(f"Channel {channel} is out of range from 0 to {self._nchannels}")

        return self._is_channel_enabled(channel)

    def set_sweep_mode(self, mode):
        if not isinstance(mode, OscilloscopeSweepMode):
            raise ValueError(f"Supplied mode {mode} is not a OscilloscopeSweepMode")

        if mode not in self._supportedSweepModes:
            raise ValueError(f"Sweep mode {mode} not supplied by device")

        self._set_sweep_mode(mode)

    def get_sweep_mode(self):
        return self._get_sweep_mode()

    def set_trigger_mode(self, mode):
        if not isinstance(mode, OscilloscopeTriggerMode):
            raise ValueError(f"Supplied mode {mode} is not a OscilloscopeTriggerMode")

        if mode not in self._supportedTriggerModes:
            raise ValueError(f"Trigger mode {mode} not supplied by device")

        self._set_trigger_mode(mode)

    def get_trigger_mode(self):
        return self._get_trigger_mode()

    def force_trigger(self):
        if self._trigger_force_supported:
            self._force_trigger()
        else:
            raise ValueError("Forcing trigger is not supported by this device")

    def set_timebase_mode(self, mode):
        if not isinstance(mode, OscilloscopeTimebaseMode):
            raise ValueError(f"Supplied mode {mode} is not a OscilloscopeTimebaseMode")
        if mode not in self._supportedTimebaseModes:
            raise ValueError(f"Timebase mode {mode} is not supported by device")
        self._set_timebase_mode(mode)

    def get_timebase_mode(self):
        return self._get_timebase_mode()

    def set_run_mode(self, mode):
        if not isinstance(mode, OscilloscopeRunMode):
            raise ValueError(f"Supplied mode {mode} is not a OscilloscopeRunMode")
        if mode not in self._supportedRunModes:
            raise ValueError(f"Run mode {mode} is not supported by device")

        self._set_run_mode(mode)

    def get_run_mode(self):
        return self._get_run_mode()

    def set_timebase_scale(self, sPerDiv):
        if not isinstance(sPerDiv, float) and not isinstance(sPerDiv, int):
            raise ValueError("Seconds per division has to be a floating point number")
        if (sPerDiv < self._timebase_scale[0]) or (sPerDiv > self._timebase_scale[1]):
            raise ValueError(f"Requested {sPerDiv}s/div is out of range of {self._timebase_scale[0]}s/div and {self._timebase_scale[1]}s/div")

        self._set_timebase_scale(sPerDiv)

    def get_timebase_scale(self):
        return self._get_timebase_scale()

    def set_channel_coupling(self, channel, couplingMode):
        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Supplied channel {channel} is out of range[0;{self._nchannels-1}]")
        if not isinstance(couplingMode, OscilloscopeCouplingMode):
            raise ValueError(f"Supplied coupling mode {couplingMode} is not a OscilloscopeCouplingMode")
        if couplingMode not in self._supportedChannelCouplingModes:
            raise ValueError(f"Coupling mode {couplingMode} is not supported by the device")

        self._set_channel_coupling(channel, couplingMode)

    def get_channel_coupling(self, channel):
        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Supplied channel {channel} is out of range [0;{self._nchannels-1}]")

        return self._get_channel_coupling(channel)

    def set_channel_probe_ratio(self, channel, ratio):
        if ratio <= 0:
            raise ValueError(f"Invalid probe ratio {ratio} supplied")
        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Supplied channel {channel} is out of range [0;{self._nchannels-1}]")
        return self._set_channel_probe_ratio(channel, ratio)

    def get_channel_probe_ratio(self, channel):
        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Supplied channel {channel} is out of range [0;{self._nchannels-1}]")
        return self._get_channel_probe_ratio(channel)

    def set_channel_scale(self, channel, scale):
        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Supplied channel {channel} is out of range [0;{self._nchannels-1}]")
        self._set_channel_scale(channel, scale)

    def get_channel_scale(self, channel):
        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Supplied channel {channel} is out of range [0;{self._nchannels-1}]")
        return self._get_channel_scale(channel)

    def query_waveform(self, channel, stats = None):
        if isinstance(channel, list) or isinstance(channel, tuple):
            # Check each supplied channel is a valid int
            for ch in channel:
                if (int(ch) < 0) or (int(ch) >= self._nchannels) or (int(ch) != ch):
                    raise ValueError(f"Supplied channel {ch} is not valid")
        else:
            channel = int(channel)
            if (channel < 0) or (channel >= self._nchannels):
                raise ValueError(f"Supplied channel {channel} is not valid")

        data = self._query_waveform(channel, stats)

        if stats is not None:
            if isinstance(stats, list) or isinstance(stats, tuple):
                for stat in stats:
                    data = self._calculate_stats(data, stat)
            else:
                # Only single statistics requested
                self._calculate_stats(data, stats)

        return data

    def _calculate_stats(self, data, stat):
        availStats = {
            "mean" : self._stats_avg,
            "fft" : self._stats_fft,
            "ifft" : self._stats_ifft,
            "correlate" : self._stats_correlate,
            "autocorrelate" : self._stats_autocorrelate
        }

        if stat in availStats:
            data = availStats[stat](data)
        return data

    def _stats_avg(self, data):
        # Do this for every channel in the data dictionary
        for iChan in range(self._nchannels):
            if f"y{iChan}" in data:
                if "means" not in data:
                    data["means"] = { }
                if (f"y{iChan}_avg" not in data) or (f"y{iChan}_std" not in data):
                    data["means"][f"y{iChan}_avg"] = np.mean(data[f"y{iChan}"])
                    data["means"][f"y{iChan}_std"] = np.std(data[f"y{iChan}"])
        return data
    def _stats_fft(self, data):
        for iChan in range(self._nchannels):
            if f"y{iChan}" in data:
                if "fft" not in data:
                    data["fft"] = { }
                if (f"y{iChan}" not in data["fft"]) or (f"y{iChan}_real" not in data["fft"]):
                    data["fft"][f"y{iChan}"] = np.fft.fft(data[f"y{iChan}"])
                    data["fft"][f"y{iChan}_real"] = np.absolute(data["fft"][f"y{iChan}"])
        return data
    def _stats_ifft(self, data):
        for iChan in range(self._nchannels):
            if f"y{iChan}" in data:
                if "ifft" not in data:
                    data["ifft"] = { }
                if f"y{iChan}" not in data["ifft"]:
                    data["ifft"][f"y{iChan}"] = np.fft.ifft(data[f"y{iChan}"])
        return data
    def _stats_autocorrelate(self, data):
        for iChan in range(self._nchannels):
            if f"y{iChan}" in data:
                if "autocorrelation" not in data:
                    data["autocorrelation"] = { }
                if f"y{iChan}" not in data["autocorrelation"]:
                    data["autocorrelation"][f"y{iChan}"] = np.correlate(data[f"y{iChan}"], data[f"y{iChan}"], mode = "full")
        return data
    def _stats_correlate(self, data):
        if self._nchannels > 1:
            for iChan1 in range(self._nchannels-1):
                for iChan2 in range(iChan1+1, self._nchannels):
                    if (f"y{iChan1}" in data) and (f"y{iChan2}" in data):
                        if "correlation" not in data:
                            data["correlation"] = { }
                        if f"y{iChan1}y{iChan2}" not in data["correlation"]:
                            data["correlation"][f"y{iChan1}y{iChan2}"] = np.correlate(data[f"y{iChan1}"], data[f"y{iChan2}"], mode = "full")
        return data

    def off(self):
        return self._off()

	# Imperative connect & disconnect methods

    def identify(self):
        return self._identify()

    def connect(self):
        if self._usesContext:
            raise ValueError("Cannot use connect on a context managed (with) object")

        res = self._connect()
        if not res:
            return res

        self._usedConnect = True
        return True

    def disconnect(self):
        if self._usesContext:
            raise ValueError("Cannot use connect on a context managed (with) object")
        if not self._usedConnect:
            return True

        self._disconnect()
        self._usedConnect = False
        return True



from enum import Enum

class GpioDirection(Enum):
    INPUT   = 0x01
    OUTPUT  = 0x02
    MASK    = 0x01|0x02

class GpioDrive(Enum):
    OPENDRAIN   = 0x04
    PUSHPULL    = 0x08
    TRISTATE    = 0x10
    MASK        = 0x04|0x08|0x10

class GpioPulls(Enum):
    PULLUP      = 0x20
    PULLDOWN    = 0x40
    MASK        = 0x20|0x40

class GPIO:
    def __init__(
        self,
        numberOfPins
    ):
        self._maxio = numberOfPins

    def getIOCount(self):
        return self._maxio

    def getConfig(self, pin):
        if (pin >= self._maxio) or (pin < 0):
            raise ValueError(f"Pin number {pin} is outside of supported range from 0 to {self._maxio-1}")

        return self._getConfig(pin)

    def setConfig(
        self,
        pin,
        name = None,
        direction = None,
        pull = None,
        drive = None,
        invertInput = False,
        invertOutput = False,
        hardwarePulsate = False
    ):
        if (pin < 0) or (pin >= self._maxio):
            raise ValueError(f"Pin number {pin} is outside of supported range from 0 to {self._maxio-1}")

        return self._setConfig(pin, name, direction, pull, drive, invertInput, invertOutput, hardwarePulsate)

    def set(self, pin, status):
        if (pin < 0) or (pin >= self._maxio):
            raise ValueError(f"Pin number {pin} outside of supported range from 0 to {self._maxio-1}")

        return self._set(pin, status)

    def get(self, pin):
        if (pin < 0) or (pin >= self._maxio):
            raise ValueError(f"Pin number {pin} outside of supported range from 0 to {self._maxio-1}")

        return self._get(pin)

    def pulse(self, pin, microseconds, state = None):
        if (pin < 0) or (pin >= self._maxio):
            raise ValueError(f"Pin number {pin} outside of supported range from 0 to {self._maxio-1}")

        return self._pulse(pin, microseconds, state)




import socket

from .exceptions import CommunicationError_Timeout, CommunicationError_NotConnected

# SCPI helper class

class SCPIDeviceEthernet:
    def __init__(self, address = None, port = 5025, logger = None):
        if not isinstance(address, str) and not (address is None):
            raise ValueError(f"Address {address} is invalid")
        if not isinstance(port, int):
            raise ValueError("Port has to be an integer value")
        if (port <= 0) or (port > 65535):
            raise ValueError("Port is out of range 1-65535")

        self._address = address
        self._port = port
        self._socket = None

    def connect(self, address = None, port = None):
        if self._socket is None:
            if address is not None:
                if not isinstance(address, str):
                    raise ValueError(f"Invalid address {address}")
                self._address = address
            if port is not None:
                if not isinstance(port, int):
                    raise ValueError("Port has to be an integer number")
                if (port <= 0) or (port > 65535):
                    raise ValueError("Port number is out of range 1-65535")
                self._port = port

            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect((self._address, self._port))
        return True

    def disconnect(self):
        if self._socket is not None:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
            self._socket = None

    def isConnected(self):
        if self._socket is not None:
            return True
        else:
            return False

    def scpiQuery(self, query):
        if not self.isConnected():
            raise CommunicationError_NotConnected("Device not connected")
        self._socket.sendall((query + "\n").encode())
        readData = ""

        while True:
            dataBlock = self._socket.recv(4096*10)
            dataBlockStr = dataBlock.decode("utf-8")
            readData = readData + dataBlockStr
            if dataBlockStr[-1] == '\n':
                break

        return readData.strip()

    def scpiCommand(self, command):
        if not self.isConnected():
            raise CommunicationError_NotConnected("Device not connected")
        self._socket.sendall((command + "\n").encode())
        return
